"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var JOI = __importStar(require("joi"));
function serialize(joi) {
    return joi.describe();
}
exports.serialize = serialize;
function str2regex(pattern) {
    var flagLen = pattern.length - pattern.lastIndexOf("/") - 1;
    return new RegExp(pattern.substr(1, pattern.length - 2 - flagLen), pattern.substr(-flagLen));
}
var list = {
    "s|string": function () { return JOI.string(); },
    "r|string:regex": function (sch, arg) { return sch.regex(arg.pattern, { name: arg.name, invert: arg.invert }); },
    "r|string:min": function (sch, arg) { return sch.min(arg); },
    "r|string:max": function (sch, arg) { return sch.max(arg); },
    "r|string:token": function (sch) { return sch.token(); },
    "r|string:alphanum": function (sch) { return sch.alphanum(); },
    "r|string:base64": function (sch) { return sch.base64(); },
    "r|string:creditCard": function (sch) { return sch.creditCard(); },
    "r|string:dataUri": function (sch) { return sch.dataUri(); },
    "r|string:email": function (sch, arg) { return sch.email(arg); },
    "r|string:guid": function (sch) { return sch.guid(); },
    "r|string:hex": function (sch, arg, desc) { return sch.hex({ byteAligned: desc.flags && desc.flags.byteAligned }); },
    "r|string:hostname": function (sch) { return sch.hostname(); },
    "r|string:ip": function (sch, arg) { return sch.ip(arg); },
    "r|string:isoDate": function (sch) { return sch.isoDate(); },
    "r|string:lowercase": function (sch) { return sch.lowercase(); },
    "r|string:length": function (sch, arg) { return sch.length(arg); },
    "r|string:uuid": function (sch) { return sch.uuid(); },
    "r|string:uppercase": function (sch) { return sch.uppercase(); },
    "r|string:uri": function (sch, arg) { return sch.uri(arg); },
    "r|string:trim": function (sch, arg) { return sch.trim(arg); },
    "r|string:normalize": function (sch, arg) { return sch.normalize(arg); },
    "f|string:case": function (sch, arg) {
        // Ignore since flag is already applied by the lowercase/uppercase rule
        /*if(arg == "lower")
            return (sch as JOI.StringSchema).lowercase() // will also add a rule
        else if(arg == "upper")
            return (sch as JOI.StringSchema).uppercase() // will also add a rule
        else
            throw new Error("Unknown value for flag 'case': "+arg);
        */
        return sch;
    },
    "f|string:insensitive": function (sch) { return sch.insensitive(); },
    "f|string:byteAligned": function (sch) { return sch; },
    "f|string:truncate": function (sch, arg) { return sch.truncate(arg); },
    "f|string:trim": function (sch) { return sch; },
    "f|string:normalize": function (sch) { return sch; },
    "s|boolean": function () { return JOI.boolean(); },
    "f|boolean:insensitive": function (sch, arg) { return sch.insensitive(arg); },
    "o|boolean:truthy": function (sch, arg) { return sch.truthy(arg.filter(function (v) { return v !== true; })); },
    "o|boolean:falsy": function (sch, arg) { return sch.falsy(arg.filter(function (v) { return v !== false; })); },
    "s|binary": function () { return JOI.binary(); },
    "f|binary:encoding": function (sch, arg) { return sch.encoding(arg); },
    "r|binary:min": function (sch, arg) { return sch.min(arg); },
    "r|binary:max": function (sch, arg) { return sch.max(arg); },
    "r|binary:length": function (sch, arg) { return sch.length(arg); },
    "s|date": function () { return JOI.date(); },
    "r|date:min": function (sch, arg) { return sch.min(arg); },
    "r|date:max": function (sch, arg) { return sch.max(arg); },
    "r|date:greater": function (sch, arg) { return sch.greater(arg); },
    "r|date:less": function (sch, arg) { return sch.less(arg); },
    "f|date:format": function (sch, arg) {
        var isoFormat = /^(?:[-+]\d{2})?(?:\d{4}(?!\d{2}\b))(?:(-?)(?:(?:0[1-9]|1[0-2])(?:\1(?:[12]\d|0[1-9]|3[01]))?|W(?:[0-4]\d|5[0-2])(?:-?[1-7])?|(?:00[1-9]|0[1-9]\d|[12]\d{2}|3(?:[0-5]\d|6[1-6])))(?![T]$|[T][\d]+Z$)(?:[T\s](?:(?:(?:[01]\d|2[0-3])(?:(:?)[0-5]\d)?|24\:?00)(?:[.,]\d+(?!:))?)(?:\2[0-5]\d(?:[.,]\d+)?)?(?:[Z]|(?:[+-])(?:[01]\d|2[0-3])(?::?[0-5]\d)?)?)?)?$/;
        if (arg.toString() != isoFormat.toString())
            throw new Error("Unknown date format!");
        return sch.iso();
    },
    "f|date:timestamp": function (sch, arg) { return sch.timestamp(arg); },
    "f|date:multiplier": function (sch) { return sch; },
    "s|object": function (sch, arg, desc) {
        if (desc.flags && desc.flags.func)
            return JOI.func();
        return JOI.object();
    },
    "o|object:children": function (sch, arg, desc, opts) {
        var obj = {};
        for (var p in arg) {
            obj[p] = deserialize(arg[p], opts);
        }
        return sch.keys(obj);
    },
    "r|object:min": function (sch, arg, desc, opts) { return sch.min(arg); },
    "r|object:max": function (sch, arg, desc, opts) { return sch.max(arg); },
    "r|object:length": function (sch, arg, desc, opts) { return sch.length(arg); },
    "o|object:patterns": function (sch, arg, desc, opts) {
        for (var _i = 0, arg_1 = arg; _i < arg_1.length; _i++) {
            var p = arg_1[_i];
            var pat = void 0;
            if (p.regex)
                pat = str2regex(p.regex);
            sch = sch.pattern(pat, deserialize(p.rule, opts));
        }
        return sch;
    },
    "o|object:dependencies": function (sch, arg, desc, opts) {
        var _a, _b;
        for (var _i = 0, arg_2 = arg; _i < arg_2.length; _i++) {
            var dep = arg_2[_i];
            switch (dep.type) {
                case 'and':
                case 'nand':
                case 'or':
                case 'xor':
                case 'oxor':
                    sch = (_a = sch)[dep.type].apply(_a, dep.peers);
                    break;
                case 'with':
                case 'without':
                    sch = (_b = sch)[dep.type].apply(_b, [dep.key].concat(dep.peers));
                    break;
                //case 'rename':
                //break;
                default:
                    console.log(dep);
                    throw new Error("Unknown object dependency:" + dep.type);
            }
        }
        return sch;
    },
    "o|object:renames": function (sch, arg, desc, opts) {
        for (var _i = 0, arg_3 = arg; _i < arg_3.length; _i++) {
            var rn = arg_3[_i];
            if (rn.isRegExp && typeof rn.from == "string")
                rn.from = str2regex(rn.from);
            sch = sch.rename(rn.from, rn.to, rn.options);
        }
        return sch;
    },
    "r|object:assert": function (sch, arg, desc, opts) { return sch.assert(JOI.ref(arg.ref.substr(4)), deserialize(arg.schema, opts)); },
    "f|object:allowUnknown": function (sch, arg, desc, opts) { return sch.unknown(arg); },
    "r|object:type": function (sch, arg, desc, opts) { return sch.type(arg.ctor, arg.name); },
    "r|object:schema": function (sch, arg, desc, opts) { return sch.schema(); },
    // Following functions are actually for .func
    "f|object:func": function (sch) { return sch; },
    "r|object:arity": function (sch, arg) { return sch.arity(arg); },
    "r|object:minArity": function (sch, arg) { return sch.minArity(arg); },
    "r|object:maxArity": function (sch, arg) { return sch.maxArity(arg); },
    "r|object:class": function (sch, arg) { return sch.class(); },
    "s|array": function () { return JOI.array(); },
    "f|array:sparse": function (sch, arg) { return sch.sparse(arg); },
    "f|array:single": function (sch, arg) { return sch.single(arg); },
    "o|array:items": function (sch, arg, desc, opts) {
        var _a;
        return (_a = sch).items.apply(_a, arg.map(function (a) { return deserialize(a, opts); }));
    },
    "o|array:orderedItems": function (sch, arg, desc, opts) {
        var _a;
        return (_a = sch).ordered.apply(_a, arg.map(function (a) { return deserialize(a, opts); }));
    },
    "r|array:min": function (sch, arg, desc, opts) { return sch.min(arg); },
    "r|array:max": function (sch, arg, desc, opts) { return sch.max(arg); },
    "r|array:length": function (sch, arg, desc, opts) { return sch.length(arg); },
    "r|array:unique": function (sch, arg, desc, opts) { return sch.unique(arg.comparator || arg.path, __assign({}, arg)); },
    "r|array:has": function (sch, arg, desc, opts) { return sch.has(deserialize(arg, opts)); },
    "s|number": function () { return JOI.number(); },
    "f|number:unsafe": function (sch, arg) { return sch.unsafe(arg); },
    "r|number:min": function (sch, arg) { return sch.min(arg); },
    "r|number:max": function (sch, arg) { return sch.max(arg); },
    "r|number:greater": function (sch, arg) { return sch.greater(arg); },
    "r|number:less": function (sch, arg) { return sch.less(arg); },
    "r|number:integer": function (sch, arg) { return sch.integer(); },
    "f|number:precision": function (sch, arg) { return sch; },
    "r|number:precision": function (sch, arg) { return sch.precision(arg); },
    "r|number:multiple": function (sch, arg) { return sch.multiple(arg); },
    "r|number:positive": function (sch, arg) { return sch.positive(); },
    "r|number:negative": function (sch, arg) { return sch.negative(); },
    "r|number:port": function (sch, arg) { return sch.port(); },
    "s|symbol": function () { return JOI.symbol(); },
    "o|symbol:map": function (sch, arg, desc, opts) {
        if (!desc.flags || !desc.flags.allowOnly)
            return sch;
        var symbolMap = opts.symbols || {};
        var arr = [];
        arg.forEach(function (k, v) {
            arr.push([k, symbolMap[k] || v]);
        });
        return sch.map(arg);
    },
    "s|alternatives": function (sch, arg, desc, opts) { return JOI.alternatives(); },
    "o|alternatives:alternatives": function (sch, arg, desc, opts) {
        var _a;
        if (arg.length == 0)
            return sch;
        arg = arg.map(function (a) { return deserialize(a, opts); });
        return (_a = sch).try.apply(_a, arg);
    },
    "s|any": function () { return JOI.any(); },
    "f|allowOnly": function (sch) { return sch; },
    "f|raw": function (sch, arg) { return sch.raw(arg); },
    "f|empty": function (sch, arg, desc) { return sch.empty(deserialize(arg)); },
    "f|error": function (sch, arg) { return sch.error(arg); },
    "f|presence": function (sch, arg) {
        if (arg == "required") {
            sch = sch.required();
        }
        else if (arg == "optional") {
            sch = sch.optional();
        }
        else if (arg == "forbidden") {
            sch = sch.forbidden();
        }
        else {
            throw new Error("invalid presence flag: " + arg);
        }
        return sch;
    },
    "f|strip": function (sch, arg) { return sch.strip(); },
    "f|default": function (sch, arg) { return sch.default(arg); },
    "o|options": function (sch, arg) { return sch.options(arg); },
    "o|meta": function (sch, arg) { return sch.meta(arg); },
    "o|examples": function (sch, arg) { return sch.example.apply(sch, arg.map(function (ex) { return [ex.value, ex.options]; })); },
    "o|description": function (sch, arg) { return sch.description(arg); },
    "o|tags": function (sch, arg) { return sch.tags(arg); },
    "o|notes": function (sch, arg) { return sch.notes(arg); },
    "o|unit": function (sch, arg) { return sch.unit(arg); },
    "o|label": function (sch, arg) { return sch.label(arg); },
    "o|invalids": function (sch, arg) { return arg.reduce(function (prev, iv) { return prev.invalid(iv); }, sch); },
    "o|valids": function (sch, arg, desc) {
        if (desc.flags && desc.flags.allowOnly)
            for (var _i = 0, arg_4 = arg; _i < arg_4.length; _i++) {
                var valid = arg_4[_i];
                sch = sch.only(valid);
            }
        else
            for (var _a = 0, arg_5 = arg; _a < arg_5.length; _a++) {
                var valid = arg_5[_a];
                sch = sch.allow(valid);
            }
        return sch;
    }
};
function applyRules(sch, desc, opts) {
    var rules = desc.rules;
    for (var _i = 0, rules_1 = rules; _i < rules_1.length; _i++) {
        var _a = rules_1[_i], arg = _a.arg, name_1 = _a.name;
        var r = list['r|' + desc.type + ":" + name_1];
        if (!r)
            r = list['r|' + name_1];
        if (!r)
            throw new Error("No rule found for " + name_1);
        sch = r(sch, arg, desc, opts);
    }
    return sch;
}
function applyFlags(sch, desc, opts) {
    var flags = desc.flags;
    for (var flag in flags) {
        var f = list['f|' + desc.type + ":" + flag];
        if (!f)
            f = list['f|' + flag];
        if (!f) {
            throw new Error("No flag found for " + flag);
        }
        sch = f(sch, flags[flag], desc, opts);
    }
    return sch;
}
function applyOpts(sch, desc, opts) {
    var options = desc;
    for (var opt in options) {
        if (['rules', 'flags', 'type'].indexOf(opt) > -1)
            continue;
        var o = list['o|' + desc.type + ":" + opt];
        if (!o)
            o = list['o|' + opt];
        if (!o)
            throw new Error("No option found for " + opt);
        sch = o(sch, options[opt], desc, opts);
    }
    return sch;
}
function deserialize(desc, opts) {
    opts = opts || {};
    var sch;
    var func = list['s|' + desc.type];
    if (!func)
        throw new Error("Unknown joi schema type: " + desc.type);
    sch = func(null, null, desc, opts);
    sch = applyOpts(sch, desc, opts);
    if (desc.rules) {
        sch = applyRules(sch, desc, opts);
    }
    if (desc.flags) {
        sch = applyFlags(sch, desc, opts);
    }
    return sch;
}
exports.deserialize = deserialize;
